from unittest import skipUnless

from django.contrib.auth.models import Permission
from django.urls import reverse
from django.utils.formats import localize
from django.utils.timezone import localtime

from cms.api import add_plugin
from cms.utils.i18n import force_language
from cms.utils.urlutils import add_url_parameters, admin_reverse

from bs4 import BeautifulSoup

from djangocms_alias.constants import (
    CHANGE_ALIAS_URL_NAME,
    USAGE_ALIAS_URL_NAME,
)
from djangocms_alias.models import (
    Alias,
    Alias as AliasModel,
    AliasContent,
    Category,
)
from djangocms_alias.utils import is_versioning_enabled
from tests.base import BaseAliasPluginTestCase


class AliasContentManagerTestCase(BaseAliasPluginTestCase):

    @skipUnless(not is_versioning_enabled(), 'Test only relevant when no versioning')
    def test_alias_content_manager_rendering_without_versioning_actions(self):
        """
        When rendering aliascontent manager without versioning
        """
        category = Category.objects.create(name='Language Filter Category')
        alias = AliasModel.objects.create(
            category=category,
            position=0,
        )
        expected_en_content = AliasContent.objects.create(
            alias=alias,
            name="EN Alias Content",
            language="en",
        )
        base_url = self.get_admin_url(AliasContent, "changelist")

        with self.login_user_context(self.superuser):
            # en is the default language configured for the site
            response = self.client.get(base_url)

        response_content_decoded = response.content.decode()

        # Check Column Headings
        self.assertInHTML(
            'Category',
            response_content_decoded,
        )

        # Check Alias content row values
        self.assertIn(
            category.name,
            response_content_decoded
        )
        self.assertIn(
            expected_en_content.name,
            response_content_decoded,
        )
        self.assertNotIn(
            expected_en_content.get_absolute_url(),
            response_content_decoded,
        )

        usage_url = admin_reverse(USAGE_ALIAS_URL_NAME, args=[expected_en_content.alias.pk])
        change_category_and_site_url = admin_reverse(
            '{}_{}_change'.format(
                expected_en_content._meta.app_label,
                expected_en_content.alias._meta.model_name
            ), args=(expected_en_content.alias.pk,)
        )

        self.assertNotIn(
            usage_url,
            response_content_decoded,
        )
        self.assertNotIn(
            change_category_and_site_url,
            response_content_decoded,
        )
        # check for add content admin link
        add_alias_link = admin_reverse(
            '{}_{}_add'.format(
                expected_en_content._meta.app_label,
                expected_en_content._meta.model_name
            )
        )
        self.assertNotIn(
            # It is not currently possible to add an alias from the django admin changelist issue #97
            # https://github.com/django-cms/djangocms-alias/issues/97
            add_alias_link,
            response_content_decoded,
        )
        self.assertNotIn(
            '<option value="delete_selected">Delete selected alias contents</option>',  # noqa: E501
            response_content_decoded
        )

    @skipUnless(is_versioning_enabled(), 'Test only relevant for versioning')
    def test_alias_changelist_rendering_with_versioning_actions(self):
        """
        When rendering aliascontent manager with versioning actions
        """
        category = Category.objects.create(name='Language Filter Category')
        alias = AliasModel.objects.create(
            category=category,
            position=0,
        )
        expected_en_content = AliasContent.objects.create(
            alias=alias,
            name="EN Alias Content",
            language="en",
        )

        from djangocms_versioning.models import Version

        Version.objects.create(content=expected_en_content, created_by=self.superuser)

        with self.login_user_context(self.superuser):

            base_url = self.get_admin_url(Alias, "changelist")
            # en is the default language configured for the site
            response = self.client.get(base_url)

        response_content_decoded = response.content.decode()

        # Check Column Headings
        self.assertInHTML(
            'Category',
            response_content_decoded,
        )
        self.assertInHTML(
            'Author',
            response_content_decoded,
        )
        self.assertInHTML(
            'Modified',
            response_content_decoded,
        )
        self.assertInHTML(
            'State',
            response_content_decoded,
        )
        self.assertInHTML(
            'Actions',
            response_content_decoded,
        )

        # Check Alias content row values
        self.assertIn(
            category.name,
            response_content_decoded
        )
        self.assertIn(
            expected_en_content.name,
            response_content_decoded,
        )

        latest_alias_content_version = expected_en_content.versions.all()[0]

        self.assertInHTML(
            f'<td class="field-get_author">{latest_alias_content_version.created_by.username}</td>',  # noqa: E501
            response_content_decoded,
        )
        self.assertIn(
            latest_alias_content_version.get_state_display(),
            response_content_decoded,
        )
        self.assertIn(
            localize(localtime(latest_alias_content_version.modified)),
            response_content_decoded,
        )

        usage_url = admin_reverse(USAGE_ALIAS_URL_NAME, args=[expected_en_content.alias.pk])
        settings_url = admin_reverse(
            '{}_{}_change'.format(
                expected_en_content._meta.app_label,
                expected_en_content.alias._meta.model_name
            ), args=(expected_en_content.alias.pk,)
        )

        self.assertIn(
            usage_url,
            response_content_decoded,
        )
        self.assertIn(
            settings_url,
            response_content_decoded,
        )

    @skipUnless(is_versioning_enabled(), 'Test only relevant for versioning')
    def test_alias_content_manager_rendering_preview_add_url(self):
        """
        When rendering aliascontent manager with versioning actions and preview
        """
        category = Category.objects.create(name='Language Filter Category')
        alias = AliasModel.objects.create(
            category=category,
            position=0,
        )
        expected_en_content = AliasContent.objects.create(
            alias=alias,
            name="EN Alias Content",
            language="en",
        )

        from djangocms_versioning.models import Version

        Version.objects.create(content=expected_en_content, created_by=self.superuser)

        with self.login_user_context(self.superuser):
            base_url = self.get_admin_url(Alias, "changelist")
            # en is the default language configured for the site
            response = self.client.get(base_url)
        response_content_decoded = response.content.decode()

        self.assertIn(
            expected_en_content.get_absolute_url(),
            response_content_decoded,
        )
        self.assertNotIn(
            '<option value="delete_selected">Delete selected alias contents</option>',  # noqa: E501
            response_content_decoded
        )
        # check for add content admin link
        add_aliascontent_url = admin_reverse(
            '{}_{}_add'.format(
                expected_en_content._meta.app_label,
                expected_en_content._meta.model_name
            )
        )
        self.assertNotIn(
            add_aliascontent_url,
            response_content_decoded,
        )

    def _create_alias_and_categories(self, category_name, alias_content_name=None):
        if not alias_content_name:
            alias_content_name = category_name
        category = Category.objects.create(name=category_name)
        alias = AliasModel.objects.create(category=category, position=0)
        alias_content = AliasContent.objects.create(
            alias=alias,
            name=alias_content_name,
            language="en"
        )
        return category, alias, alias_content

    @skipUnless(is_versioning_enabled(), 'Test only relevant for versioning')
    def test_category_field_ordering_versioned(self):
        """
        Related category can be ordered by name, both in ascending and descending order, with versioning
        """
        # Create a number of categories, aliases, and alias content to order
        first_category, first_alias, first_alias_content = self._create_alias_and_categories("First Order Test Case")
        # Previously lowercase and upper case would be sorted separately, test they are ordered together
        first_category_lower, first_alias_lower, first_alias_content_lower = self._create_alias_and_categories(
            "first order test case lower"
        )
        middle_category, middle_alias, middle_alias_content = self._create_alias_and_categories(
            "Middle Order Test Case"
        )
        # Previously lowercase and upper case would be sorted separately, test they are ordered together
        last_category_lower, last_alias_lower, last_alias_content_lower = self._create_alias_and_categories(
            "z order test case lower"
        )
        last_category, last_alias, last_alias_content = self._create_alias_and_categories(
            "Z Order Test Case Upper"
        )
        # Create the versions for each alias content
        from djangocms_versioning.models import Version
        Version.objects.create(content=first_alias_content, created_by=self.superuser)
        Version.objects.create(content=first_alias_content_lower, created_by=self.superuser)
        Version.objects.create(content=middle_alias_content, created_by=self.superuser)
        Version.objects.create(content=last_alias_content_lower, created_by=self.superuser)
        Version.objects.create(content=last_alias_content, created_by=self.superuser)

        with self.login_user_context(self.superuser):
            base_url = self.get_admin_url(Alias, "changelist")
            # o=1 indicates ascending alphabetical order on list_displays second entry
            base_url += "?o=1"
            # en is the default language configured for the site
            response = self.client.get(base_url)
        soup = BeautifulSoup(response.content, "html.parser")
        results = soup.find_all("td", class_="field-category")

        # Test results are in ascending alphabetical order
        self.assertEqual(results[0].text, first_alias_content.name)
        self.assertEqual(results[1].text, first_alias_content_lower.name)
        self.assertEqual(results[2].text, middle_alias_content.name)
        self.assertEqual(results[3].text, last_alias_content_lower.name)
        self.assertEqual(results[4].text, last_alias_content.name)

        with self.login_user_context(self.superuser):
            base_url = self.get_admin_url(Alias, "changelist")
            # o=-1 indicates descending alphabetical order on list_displays second entry
            base_url += "?o=-1"
            # en is the default language configured for the site
            response = self.client.get(base_url)
        soup = BeautifulSoup(response.content, "html.parser")
        results = soup.find_all("td", class_="field-category")
        # Test results are in descending alphabetical order
        self.assertEqual(results[4].text, first_alias_content.name)
        self.assertEqual(results[3].text, first_alias_content_lower.name)
        self.assertEqual(results[2].text, middle_alias_content.name)
        self.assertEqual(results[1].text, last_alias_content_lower.name)
        self.assertEqual(results[0].text, last_alias_content.name)

    @skipUnless(not is_versioning_enabled(), 'Test only relevant for versioning')
    def test_category_field_ordering_unversioned(self):
        """
        Related category can be ordered by name, both in ascending and descending order, without versioning
        """
        # Create a number of categories, aliases, and alias content to order
        first_category, first_alias, first_alias_content = self._create_alias_and_categories("First Order Test Case")
        # Previously lowercase and upper case would be sorted separately, test they are ordered together
        first_category_lower, first_alias_lower, first_alias_content_lower = self._create_alias_and_categories(
            "first order test case lower"
        )
        middle_category, middle_alias, middle_alias_content = self._create_alias_and_categories(
            "Middle Order Test Case"
        )
        # Previously lowercase and upper case would be sorted separately, test they are ordered together
        last_category_lower, last_alias_lower, last_alias_content_lower = self._create_alias_and_categories(
            "z order test case lower"
        )
        last_category, last_alias, last_alias_content = self._create_alias_and_categories(
            "Z Order Test Case Upper"
        )

        with self.login_user_context(self.superuser):
            base_url = self.get_admin_url(AliasContent, "changelist")
            # o=1 indicates ascending alphabetical order on list_displays second entry
            base_url += "?o=1"
            # en is the default language configured for the site
            response = self.client.get(base_url)
        soup = BeautifulSoup(response.content, "html.parser")
        results = soup.find_all("td", class_="field-get_category")

        # Test results are in ascending alphabetical order
        self.assertEqual(results[0].text, first_alias_content.name)
        self.assertEqual(results[1].text, first_alias_content_lower.name)
        self.assertEqual(results[2].text, middle_alias_content.name)
        self.assertEqual(results[3].text, last_alias_content_lower.name)
        self.assertEqual(results[4].text, last_alias_content.name)

        with self.login_user_context(self.superuser):
            base_url = self.get_admin_url(AliasContent, "changelist")
            # o=-1 indicates descending alphabetical order on list_displays second entry
            base_url += "?o=-1"
            # en is the default language configured for the site
            response = self.client.get(base_url)
        soup = BeautifulSoup(response.content, "html.parser")
        results = soup.find_all("td", class_="field-get_category")

        # Test results are in descending alphabetical order
        self.assertEqual(results[4].text, first_alias_content.name)
        self.assertEqual(results[3].text, first_alias_content_lower.name)
        self.assertEqual(results[2].text, middle_alias_content.name)
        self.assertEqual(results[1].text, last_alias_content_lower.name)
        self.assertEqual(results[0].text, last_alias_content.name)

    def test_aliascontent_list_view(self):
        """
        The AliasContent admin change list displays correct details
        about the objects
        """
        category1 = Category.objects.create(
            name='Category 1',
        )
        category2 = Category.objects.create(
            name='Category 2',
        )

        plugin = add_plugin(
            self.placeholder,
            'TextPlugin',
            language=self.language,
            body='This is basic content',
        )

        alias1 = self._create_alias(
            [plugin],
            name='Alias 1',
            category=category1,
        )
        alias2 = self._create_alias(
            [plugin],
            name='Alias 2',
            category=category2,
        )
        alias3 = self._create_alias(
            [plugin],
            name='Alias 3',
            category=category1,
            published=False,
        )

        with self.login_user_context(self.superuser):
            response = self.client.get(
                self.get_list_alias_endpoint(),
            )

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, category1.name)
        self.assertContains(response, category2.name)
        self.assertContains(response, 'Alias 1')
        self.assertContains(response, 'Alias 2')
        self.assertContains(response, 'Alias 3')

        if is_versioning_enabled():
            # we have both published and draft aliases so both should
            # be displayed
            self.assertContains(response, "Published")
            self.assertContains(response, "Draft")
        else:
            self.assertNotContains(response, "Published")
            self.assertNotContains(response, "Draft")

        aliascontent1_url = alias1.get_absolute_url()
        aliascontent2_url = alias2.get_absolute_url()
        aliascontent3_url = alias3.get_absolute_url()

        # when versioning is not enabled, the django admin change form
        # is used which used links to the aliascontent_change view
        if not is_versioning_enabled():
            alias1_content = alias1.get_content(language=self.language)
            alias2_content = alias2.get_content(language=self.language)
            alias3_content = alias3.get_content(language=self.language)
            aliascontent1_url = admin_reverse(
                CHANGE_ALIAS_URL_NAME, args=[alias1_content.pk]
            )
            aliascontent2_url = admin_reverse(
                CHANGE_ALIAS_URL_NAME, args=[alias2_content.pk]
            )
            aliascontent3_url = admin_reverse(
                CHANGE_ALIAS_URL_NAME, args=[alias3_content.pk]
            )

        self.assertContains(response, aliascontent1_url)
        self.assertContains(response, aliascontent2_url)
        self.assertContains(response, aliascontent3_url)


class CategoryAdminViewsTestCase(BaseAliasPluginTestCase):

    def test_changelist(self):
        Category.objects.all().delete()
        category1 = Category.objects.create()
        category2 = Category.objects.create()
        category1.translations.create(language_code='en', name='Category 1')
        category2.translations.create(language_code='en', name='Category 2')
        category1.translations.create(language_code='de', name='Kategorie 1')
        category2.translations.create(language_code='fr', name='Catégorie 2')
        category1.translations.create(language_code='it', name='Categoria 1')

        with self.login_user_context(self.superuser):
            with force_language('en'):
                en_response = self.client.get(self.get_category_list_endpoint())
            with force_language('de'):
                de_response = self.client.get(self.get_category_list_endpoint())
            with force_language('fr'):
                fr_response = self.client.get(self.get_category_list_endpoint())
            with force_language('it'):
                it_response = self.client.get(self.get_category_list_endpoint())

        self.assertContains(en_response, 'Category 1')
        self.assertContains(en_response, 'Category 2')
        self.assertNotContains(en_response, 'Kategorie 1')
        self.assertNotContains(en_response, 'Catégorie 2')
        self.assertNotContains(en_response, 'Categoria 1')

        self.assertContains(de_response, 'Kategorie 1')
        self.assertContains(de_response, 'Category 2')  # fallback
        self.assertNotContains(de_response, 'Category 1')
        self.assertNotContains(de_response, 'Catégorie 2')
        self.assertNotContains(de_response, 'Categoria 1')

        self.assertContains(fr_response, 'Category 1')  # fallback
        self.assertContains(fr_response, 'Catégorie 2')
        self.assertNotContains(fr_response, 'Category 2')
        self.assertNotContains(fr_response, 'Kategorie 1')
        self.assertNotContains(fr_response, 'Categoria 2')

        self.assertContains(it_response, 'Catégorie 2')  # fallback
        self.assertNotContains(it_response, 'Category 1')
        self.assertNotContains(it_response, 'Category 2')
        self.assertNotContains(it_response, 'Kategorie 1')
        self.assertNotContains(it_response, 'Categoria 2')

    def test_changelist_standard_user(self):
        """
        Expect a 302 redirect as the view is handled by the django admin
        """
        with self.login_user_context(self.get_standard_user()):
            response = self.client.get(self.get_category_list_endpoint())
        self.assertEqual(response.status_code, 302)

    def test_changelist_staff_user_without_permission(self):
        url = self.get_category_list_endpoint()
        with self.login_user_context(self.get_staff_user_with_std_permissions()):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 403)

    def test_changelist_staff_user_with_permission(self):
        user = self.get_staff_user_with_std_permissions()
        user.user_permissions.add(Permission.objects.get(
            content_type__app_label='djangocms_alias',
            codename='change_category'))
        with self.login_user_context(user):
            response = self.client.get(self.get_category_list_endpoint())
        self.assertEqual(response.status_code, 200)

    def test_changelist_edit_button(self):
        with self.login_user_context(self.superuser):
            response = self.client.get(self.get_category_list_endpoint())

        self.assertContains(
            response,
            '<a href="/en/admin/djangocms_alias/category/1/change/"'
        )

    def test_change_view(self):
        with self.login_user_context(self.superuser):
            self.client.post(
                add_url_parameters(
                    admin_reverse(
                        'djangocms_alias_category_change',
                        args=[self.category.pk],
                    ),
                    language='de',
                ),
                data={
                    'name': 'Alias Kategorie',
                },
            )

        self.assertEqual(self.category.name, 'test category')
        self.category.set_current_language('de')
        self.assertEqual(self.category.name, 'Alias Kategorie')


class AliasesManagerTestCase(BaseAliasPluginTestCase):

    def test_aliases_admin_entry_is_hidden(self):
        """
        Aliases admin entry should not be available via the admin menu
        """
        index_url = reverse('admin:index')

        self.client.force_login(self.superuser)

        response = self.client.get(index_url)

        unexpected_content = '<th scope="row"><a href="/en/admin/djangocms_alias/aliascontent/">Alias contents</a></th>'
        expected_content = '<th scope="row"><a href="/en/admin/djangocms_alias/alias/">Aliases</a></th>'

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, expected_content)
        self.assertNotContains(response, unexpected_content)

    def test_aliases_endpoint_accessible_via_url(self):
        """
        Aliases admin endpoint should still be accessible via the endpoint
        """
        base_url = self.get_admin_url(AliasModel, "changelist")

        with self.login_user_context(self.superuser):
            response = self.client.get(base_url)

        module_name = response.context_data['module_name']

        self.assertEqual(response.status_code, 200)
        self.assertEqual(module_name, 'aliases')



from django.contrib.auth import get_permission_codename
from django.contrib.auth.models import Permission
from django.http import QueryDict
from django.test.client import RequestFactory

from cms.api import add_plugin, create_page, create_page_content
from cms.middleware.toolbar import ToolbarMiddleware
from cms.test_utils.testcases import CMSTestCase
from cms.toolbar.utils import (
    get_object_edit_url,
    get_object_preview_url,
    get_object_structure_url,
)
from cms.utils.conf import get_cms_setting
from cms.utils.urlutils import admin_reverse

from djangocms_alias.constants import (
    CATEGORY_LIST_URL_NAME,
    CREATE_ALIAS_URL_NAME,
    DELETE_ALIAS_URL_NAME,
    DETACH_ALIAS_PLUGIN_URL_NAME,
    LIST_ALIAS_URL_NAME,
)
from djangocms_alias.models import Alias as AliasModel, AliasContent, Category
from djangocms_alias.utils import is_versioning_enabled


class BaseAliasPluginTestCase(CMSTestCase):

    def get_create_alias_endpoint(self):
        return admin_reverse(CREATE_ALIAS_URL_NAME)

    def get_category_list_endpoint(self):
        return admin_reverse(CATEGORY_LIST_URL_NAME)

    def get_detach_alias_plugin_endpoint(self, plugin_pk):
        return admin_reverse(
            DETACH_ALIAS_PLUGIN_URL_NAME,
            args=[plugin_pk],
        )

    def get_delete_alias_endpoint(self, alias_pk):
        return admin_reverse(
            DELETE_ALIAS_URL_NAME,
            args=[alias_pk],
        )

    def get_list_alias_endpoint(self):
        return admin_reverse(LIST_ALIAS_URL_NAME)

    def setUp(self):
        self.superuser = self.get_superuser()
        self.language = 'en'
        self.page = self._create_page('test')
        self.placeholder = self.page.get_placeholders(self.language).get(
            slot='content',
        )
        self.plugin = add_plugin(
            self.placeholder,
            'TextPlugin',
            language=self.language,
            body='test',
        )
        self.category = Category.objects.create(name='test category')

    def _get_draft_page_placeholder(self):
        page_content = create_page_content(self.language, 'Draft Page', self.page, created_by=self.superuser)
        return page_content.get_placeholders().get(slot='content')

    def _create_alias(self, plugins=None, name='test alias', category=None, position=0,
                      language=None, published=True, static_code=None, site=None):
        if language is None:
            language = self.language
        if category is None:
            category = self.category
        if plugins is None:
            plugins = []
        alias = AliasModel.objects.create(
            category=category,
            position=position,
            static_code=static_code,
            site=site,
        )
        alias_content = AliasContent.objects.create(
            alias=alias,
            name=name,
            language=language,
        )

        if is_versioning_enabled():
            from djangocms_versioning.models import Version
            version = Version.objects.create(content=alias_content, created_by=self.superuser)
            if published:
                version.publish(self.superuser)

        if plugins:
            alias_content.populate(plugins=plugins)
        return alias

    def _get_version(self, grouper, version_state, language=None):
        language = language or self.language

        from djangocms_versioning.models import Version
        versions = Version.objects.filter_by_grouper(grouper).filter(state=version_state)
        for version in versions:
            if hasattr(version.content, 'language') and version.content.language == language:
                return version

    def _publish(self, grouper, language=None):
        from djangocms_versioning.constants import DRAFT
        version = self._get_version(grouper, DRAFT, language)
        version.publish(self.superuser)

    def _unpublish(self, grouper, language=None):
        from djangocms_versioning.constants import PUBLISHED
        version = self._get_version(grouper, PUBLISHED, language)
        version.unpublish(self.superuser)

    def _create_page(self, title, language=None, site=None, published=True, **kwargs):
        if language is None:
            language = self.language

        if is_versioning_enabled() and not kwargs.get('created_by'):
            kwargs['created_by'] = self.superuser

        page = create_page(
            title=title,
            language=language,
            template='page.html',
            menu_title='',
            in_navigation=True,
            limit_visibility_in_menu=None,
            site=site,
            **kwargs
        )
        if is_versioning_enabled() and published:
            self._publish(page, language)
        return page

    def get_alias_request(self, alias, lang_code='en', *args, **kwargs):
        request = self._get_instance_request(alias, *args, **kwargs)
        request.current_page = None
        request = self._process_request_by_toolbar_middleware(request, obj=alias.get_content(lang_code))
        return request

    def get_page_request(self, page, obj=None, *args, **kwargs):
        request = self._get_instance_request(page, *args, **kwargs)
        request.current_page = page
        request = self._process_request_by_toolbar_middleware(request, obj)
        return request

    def _get_instance_request(self, instance, user, path=None, edit=False,
                              preview=False, structure=False, lang_code='en',
                              disable=False):
        if not path:
            if edit:
                path = get_object_edit_url(instance)
            elif preview:
                path = get_object_preview_url(instance)
            elif structure:
                path = get_object_structure_url(instance)
            else:
                path = instance.get_absolute_url()

        request = RequestFactory().get(path)
        request.session = {}
        request.user = user
        request.LANGUAGE_CODE = lang_code
        request.GET = QueryDict('', mutable=True)
        if edit:
            request.GET['edit'] = None
        else:
            request.GET['edit_off'] = None
        if disable:
            request.GET[get_cms_setting('CMS_TOOLBAR_URL__DISABLE')] = None

        return request

    def _process_request_by_toolbar_middleware(self, request, obj=None):
        middleware = ToolbarMiddleware(request)
        middleware.process_request(request)
        if hasattr(request, 'toolbar'):
            if obj:
                request.toolbar.set_object(obj)
            request.toolbar.populate()
            request.toolbar.post_template_populate()
        return request

    def _add_default_permissions(self, user):
        # Text plugin permissions
        user.user_permissions.add(Permission.objects.get(codename='add_text'))
        user.user_permissions.add(Permission.objects.get(codename='delete_text'))
        user.user_permissions.add(Permission.objects.get(codename='change_text'))
        user.user_permissions.add(Permission.objects.get(codename='add_page'))
        user.user_permissions.add(Permission.objects.get(codename='change_page'))
        user.user_permissions.add(Permission.objects.get(codename='delete_page'))

    def add_alias_plugin_to_page(self, page, alias, language=None):
        if language is None:
            language = self.language

        add_plugin(
            page.get_placeholders(language).get(slot='content'),
            'Alias',
            language=language,
            alias=alias,
        )

    def get_staff_user_with_alias_permissions(self):
        staff_user = self._create_user("alias staff", is_staff=True, is_superuser=False)  # noqa: E501
        self.add_permission(staff_user, get_permission_codename('add', AliasModel._meta))  # noqa: E501
        self.add_permission(staff_user, get_permission_codename('change', AliasModel._meta))  # noqa: E501
        self.add_permission(staff_user, get_permission_codename('delete', AliasModel._meta))  # noqa: E501
        self.add_permission(staff_user, get_permission_codename('add', AliasContent._meta))  # noqa: E501
        self.add_permission(staff_user, get_permission_codename('change', AliasContent._meta))  # noqa: E501
        self.add_permission(staff_user, get_permission_codename('delete', AliasContent._meta))  # noqa: E501
        self.add_permission(staff_user, get_permission_codename('add', Category._meta))  # noqa: E501
        return staff_user



import itertools
from collections import ChainMap
from unittest import skipUnless

from django.contrib.auth.models import Permission
from django.urls import reverse

from cms.cms_toolbars import (
    ADMIN_MENU_IDENTIFIER,
    ADMINISTRATION_BREAK,
    LANGUAGE_MENU_IDENTIFIER,
)
from cms.toolbar.items import Break, ButtonList, ModalItem
from cms.toolbar.utils import get_object_edit_url
from cms.utils.i18n import force_language
from cms.utils.urlutils import admin_reverse

from djangocms_alias.cms_toolbars import ALIAS_MENU_IDENTIFIER
from djangocms_alias.constants import USAGE_ALIAS_URL_NAME
from djangocms_alias.models import AliasContent
from djangocms_alias.utils import is_versioning_enabled

from .base import BaseAliasPluginTestCase


class AliasToolbarTestCase(BaseAliasPluginTestCase):

    def _get_wizard_create_button(self, request):
        button_lists = [
            result.item
            for result in request.toolbar.find_items(item_type=ButtonList)
        ]
        buttons = list(
            # flatten the list
            itertools.chain.from_iterable([
                item.buttons
                for item in button_lists
            ])
        )
        # There will always be this button, because we are in the context of
        # alias app views
        return [
            button for button in buttons if button.name == 'Create'
        ][0]

    def test_add_aliases_submenu_to_admin_menu_no_permission(self):
        with self.login_user_context(self.get_staff_user_with_std_permissions()):
            response = self.client.get(self.page.get_absolute_url())
        self.assertNotContains(response, '<span>Aliases')

    def test_add_aliases_submenu_to_admin_menu(self):
        user = self.get_staff_user_with_std_permissions()
        user.user_permissions.add(Permission.objects.get(
            content_type__app_label='djangocms_alias',
            codename='change_category'))
        try:
            page_url = get_object_edit_url(self.page.get_title_obj(self.language))
        except AttributeError:
            page_url = get_object_edit_url(self.page.get_content_obj(self.language))
        with self.login_user_context(user):
            response = self.client.get(page_url, follow=True)
        self.assertContains(response, '<span>Aliases')

    def test_aliases_link_placement(self):
        request = self.get_page_request(self.page, user=self.superuser)
        admin_menu = request.toolbar.get_or_create_menu(ADMIN_MENU_IDENTIFIER)
        break_item = admin_menu.find_first(Break, identifier=ADMINISTRATION_BREAK)  # noqa: E501
        item_positioned_before_admin_break = admin_menu.items[break_item.index - 1]  # noqa: E501
        self.assertEqual(item_positioned_before_admin_break.name, 'Aliases...')

    def test_add_alias_menu_showing_only_on_alias_plugin_views(self):
        alias = self._create_alias([self.plugin])
        for endpoint in [
            self.get_category_list_endpoint(),
            self.get_list_alias_endpoint(),
            self.page.get_absolute_url(language=self.language),
        ]:
            request = self.get_page_request(page=None, path=endpoint, user=self.superuser)
            alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)
            self.assertEqual(alias_menu, None)

        def _test_alias_endpoint(**kwargs):
            kwargs.update({
                'alias': alias,
                'path': endpoint,
                'user': self.superuser,
            })
            # py34 compat
            request = self.get_alias_request(**ChainMap(kwargs))
            alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)
            self.assertEqual(alias_menu.name, 'Alias')

        _test_alias_endpoint()
        _test_alias_endpoint(edit=True)
        _test_alias_endpoint(preview=True)

    def test_alias_toolbar_language_menu(self):
        request = self.get_page_request(self.page, user=self.superuser)
        alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)
        self.assertEqual(alias_menu, None)

        alias = self._create_alias([self.plugin])

        request = self.get_page_request(
            page=None,
            path=self.get_category_list_endpoint(),
            user=self.superuser,
            edit=True,
        )
        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
        self.assertEqual(language_menu.get_item_count(), 4)

        request = self.get_alias_request(
            alias=alias,
            user=self.superuser,
            preview=True,
        )
        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
        self.assertEqual(language_menu.get_item_count(), 1)

        request = self.get_alias_request(
            alias=alias,
            user=self.superuser,
            edit=True,
        )
        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
        self.assertEqual(language_menu.get_item_count(), 4)

        language_menu_dict = {
            menu.name: [menu_item.name for menu_item in menu.items]
            for key, menu in language_menu.menus.items()
        }
        self.assertIn('Add Translation', language_menu_dict.keys())
        self.assertIn('Delete Translation', language_menu_dict.keys())
        self.assertEqual(
            {'Deutsche...', 'Française...', 'Italiano...'},
            set(language_menu_dict['Add Translation']),
        )
        self.assertEqual(
            {'English...'},
            set(language_menu_dict['Delete Translation']),
        )

        alias_content = alias.contents.create(name='test alias 2', language='fr')
        alias_content.populate(replaced_placeholder=self.placeholder)
        alias_content.alias.clear_cache()

        if is_versioning_enabled():
            from djangocms_versioning.constants import PUBLISHED
            from djangocms_versioning.models import Version

            Version.objects.create(
                content=alias_content, created_by=self.superuser, state=PUBLISHED)

        request = self.get_alias_request(
            alias=alias,
            user=self.superuser,
            edit=True,
        )
        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
        self.assertEqual(language_menu.get_item_count(), 6)

        language_menu_dict = {
            menu.name: [menu_item.name for menu_item in menu.items]
            for key, menu in language_menu.menus.items()
        }
        self.assertEqual(
            {'Deutsche...', 'Italiano...'},
            set(language_menu_dict['Add Translation']),
        )
        self.assertEqual(
            {'Française...', 'English...'},
            set(language_menu_dict['Delete Translation']),
        )
        self.assertEqual(
            {'from Française'},
            set(language_menu_dict['Copy all plugins']),
        )
        language_menu_first_items = {
            menu.name: next(filter(
                lambda item: item.name in ['Française...', 'Deutsche...', 'from Française'],
                menu.items,
            ))
            for key, menu in language_menu.menus.items()
        }
        # First item is Deutsche... for Add Translation
        self.assertIn(
            '/en/admin/djangocms_alias/aliascontent/add/',
            language_menu_first_items['Add Translation'].url,
        )
        self.assertIn(
            'language=de',
            language_menu_first_items['Add Translation'].url,
        )
        self.assertIn(
            f'alias={alias.pk}',
            language_menu_first_items['Add Translation'].url,
        )
        self.assertEqual(
            # First item is Française... for Delete Translation
            '/en/admin/djangocms_alias/aliascontent/{}/delete/?language=fr'.format(
                alias.get_content('fr').pk,
            ),
            language_menu_first_items['Delete Translation'].url,
        )
        self.assertRegex(
            language_menu_first_items['Copy all plugins'].action,
            r'en\/admin\/([\w\/]+)\/copy-plugins\/',
        )

    def test_language_switcher_when_toolbar_object_is_alias_content(self):
        alias = self._create_alias([self.plugin])
        alias_content = alias.contents.create(name='test alias 2', language='fr')
        expected_result = ['English', 'Française']

        if is_versioning_enabled():
            from djangocms_versioning.constants import DRAFT
            from djangocms_versioning.models import Version
            Version.objects.create(
                content=alias_content, created_by=self.superuser, state=DRAFT)

        alias_content.populate(replaced_placeholder=self.placeholder)
        alias_content.alias.clear_cache()
        request = self.get_alias_request(
            alias=alias,
            user=self.superuser,
            preview=True,
        )
        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)

        self.assertEqual([item.name for item in language_menu.items], expected_result)

    def test_language_switcher_when_toolbar_object_isnt_alias_content(self):
        request = self.get_page_request(
            page=self.page,
            user=self.superuser,
            preview=True,
        )
        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
        expected_result = ['English', 'Deutsche', 'Française', 'Italiano']
        # Versioning changes the toolbar language selector and only shows
        # languages that have translations
        if is_versioning_enabled():
            expected_result = ['English']

        # Don't change default language switcher that is used for Pages
        self.assertEqual(
            [item.name for item in language_menu.items], expected_result
        )

    def test_page_toolbar_no_language_menu(self):
        from django.utils.translation import gettext as _

        alias = self._create_alias([self.plugin])
        alias_content = alias.contents.create(name='test alias 2', language='fr')
        # Get request
        request = self.get_alias_request(
            alias=alias,
            lang_code="fr",
            path=get_object_edit_url(alias_content),
            user=self.get_superuser(),
        )
        # Remove language menu from request's toolbar
        del request.toolbar.menus[LANGUAGE_MENU_IDENTIFIER]

        # find VersioningPageToolbar
        for cls, toolbar in request.toolbar.toolbars.items():
            if cls == "djangocms_alias.cms_toolbars.AliasToolbar":
                # and call override_language_menu
                toolbar.override_language_switcher()
                break
        else:
            self.fail("No AliasToolbar in alias request")
        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER, _("Language"))
        self.assertIsNone(language_menu)

    def test_change_alias_settings_button_is_visible_on_alias_edit_view(self):
        button_label = 'Change alias settings...'
        alias_change_viewname = 'djangocms_alias_alias_change'
        alias = self._create_alias()
        with force_language('en'):
            request = self.get_alias_request(
                alias=alias,
                user=self.superuser,
                edit=True,
            )
        alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)
        search_result = alias_menu.find_first(item_type=ModalItem, name=button_label)
        self.assertIsNotNone(search_result)
        button = search_result.item
        self.assertEqual(button.on_close, 'REFRESH_PAGE')
        self.assertEqual(
            button.url,
            admin_reverse(
                alias_change_viewname,
                args=[alias.pk],
            ),
        )

    def test_alias_usage_button(self):
        alias = self._create_alias()
        request = self.get_alias_request(
            alias=alias,
            user=self.superuser,
            edit=True,
        )
        button_label = 'View usage...'
        alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)
        search_result = alias_menu.find_first(item_type=ModalItem, name=button_label)
        self.assertIsNotNone(search_result)
        button = search_result.item
        self.assertEqual(
            button.url,
            admin_reverse(
                USAGE_ALIAS_URL_NAME,
                args=[alias.pk],
            ),
        )
        self.assertEqual(
            button.on_close,
            'REFRESH_PAGE',
        )

    def test_create_wizard_button_enabled(self):
        request = self.get_page_request(
            page=None,
            path=self.get_category_list_endpoint(),
            user=self.superuser,
        )
        create_button = self._get_wizard_create_button(request)
        self.assertEqual(create_button.disabled, False)

    @skipUnless(not is_versioning_enabled(), 'Test only relevant when no versioning')
    def test_delete_button_show_on_edit_alias_view_no_versioning(self):
        """
        When versioning is not installed deletion should be possible. The delete button
        should be available in the toolbar.
        """
        alias = self._create_alias()
        request = self.get_alias_request(
            alias=alias,
            user=self.superuser,
            edit=True,
        )
        button_label = 'Delete alias...'
        alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)

        search_result = alias_menu.find_first(item_type=ModalItem, name=button_label)

        self.assertIsNotNone(search_result)

        button = search_result.item

        self.assertEqual(button.name, button_label)
        self.assertEqual(button.url, self.get_delete_alias_endpoint(alias.pk))
        self.assertEqual(
            button.on_close,
            self.get_list_alias_endpoint(),
        )

    @skipUnless(is_versioning_enabled(), 'Test only relevant for versioning')
    def test_delete_button_not_shown_on_edit_alias_view_with_versioning(self):
        """
        When versioning is installed no deletion should be possible. The delete button
        should not be available in the toolbar.
        """
        alias = self._create_alias()
        request = self.get_alias_request(
            alias=alias,
            user=self.superuser,
            edit=True,
        )
        button_label = 'Delete alias...'
        alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)
        search_result = alias_menu.find_first(item_type=ModalItem, name=button_label)

        # No button should be found for delete
        self.assertIsNone(search_result)

    def test_do_not_disable_buttons_when_in_preview_mode(self):
        alias = self._create_alias()
        request = self.get_alias_request(
            alias=alias,
            user=self.superuser,
            preview=True,
        )
        alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)
        search_results = alias_menu.find_items(item_type=ModalItem)
        self.assertNotEqual(bool(search_results), False)
        for result in search_results:
            if result.item.name == 'View usage...':
                self.assertEqual(result.item.disabled, False)
            else:
                self.assertEqual(result.item.disabled, False)

    def test_disable_buttons_when_not_have_perms(self):
        alias = self._create_alias()
        staff_user = self._create_user("user1", is_staff=True, is_superuser=False)
        request = self.get_alias_request(
            alias=alias,
            user=staff_user,
            edit=True,
        )
        alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)
        search_results = alias_menu.find_items(item_type=ModalItem)
        self.assertNotEqual(bool(search_results), False)
        for result in search_results:
            if result.item.name == 'View usage...':
                self.assertEqual(result.item.disabled, False)
            else:
                self.assertEqual(result.item.disabled, True)

    def test_enable_buttons_when_on_edit_mode(self):
        alias = self._create_alias()
        request = self.get_alias_request(
            alias=alias,
            user=self.superuser,
            edit=True,
        )
        alias_menu = request.toolbar.get_menu(ALIAS_MENU_IDENTIFIER)
        search_results = alias_menu.find_items(item_type=ModalItem)
        self.assertNotEqual(bool(search_results), False)
        for result in search_results:
            self.assertEqual(result.item.disabled, False)

    def test_site_dropdown_url_renders_admin_changelist(self):
        request = self.get_page_request(self.page, user=self.superuser)
        admin_menu = request.toolbar.get_or_create_menu(ADMIN_MENU_IDENTIFIER)
        site_aliases_url = admin_menu.items[3].url

        with self.login_user_context(self.superuser):
            response = self.client.get(
                site_aliases_url,
            )

        content = response.content.decode('utf-8')

        self.assertEqual(response.status_code, 200)
        # Rendered content should contain admin changelist header
        self.assertIn("Select alias to change | Django site admin", content)

    def test_site_dropdown_url_renders_admin_changelist_url(self):
        request = self.get_page_request(self.page, user=self.superuser)
        admin_menu = request.toolbar.get_or_create_menu(ADMIN_MENU_IDENTIFIER)
        site_aliases_url = admin_menu.items[3].url
        admin_changelist_aliases_url = reverse("admin:{}_alias_changelist".format(
            AliasContent._meta.app_label)
        ) + "?language=en"

        with self.login_user_context(self.superuser):
            response = self.client.get(
                site_aliases_url,
            )

        self.assertEqual(response.status_code, 200)
        # Url from site menu renders Alias content
        self.assertEqual(site_aliases_url, admin_changelist_aliases_url)



import itertools
from copy import copy

from django.urls import NoReverseMatch
from django.utils.encoding import force_str
from django.utils.http import urlencode
from django.utils.translation import (
    get_language_from_request,
    gettext,
    gettext_lazy as _,
)

from cms.cms_toolbars import (
    ADMIN_MENU_IDENTIFIER,
    ADMINISTRATION_BREAK,
    LANGUAGE_MENU_IDENTIFIER,
    SHORTCUTS_BREAK,
)
from cms.toolbar.items import Break, ButtonList
from cms.toolbar_base import CMSToolbar
from cms.toolbar_pool import toolbar_pool
from cms.utils.i18n import (
    force_language,
    get_default_language,
    get_language_dict,
    get_language_tuple,
)
from cms.utils.permissions import get_model_permission_codename
from cms.utils.urlutils import add_url_parameters, admin_reverse

from .constants import (
    DELETE_ALIAS_URL_NAME,
    LIST_ALIAS_URL_NAME,
    USAGE_ALIAS_URL_NAME,
)
from .models import Alias, AliasContent
from .utils import is_versioning_enabled


__all__ = [
    'AliasToolbar',
]


ALIAS_MENU_IDENTIFIER = 'alias'
ALIAS_MENU_CREATE_IDENTIFIER = 'alias-add'
ALIAS_LANGUAGE_BREAK = 'alias-language'


@toolbar_pool.register
class AliasToolbar(CMSToolbar):
    name = _('Alias')
    plural_name = _('Aliases')

    def populate(self):
        self.add_aliases_link_to_admin_menu()

        if isinstance(self.toolbar.obj, AliasContent):
            self.add_alias_menu()
            self.override_language_switcher()
            self.change_language_menu()

    def post_template_populate(self):
        if self.is_current_app or isinstance(self.toolbar.obj, AliasContent):
            self.enable_create_wizard_button()

    def add_aliases_link_to_admin_menu(self):
        if not self.request.user.has_perm('djangocms_alias.change_category'):
            return
        admin_menu = self.toolbar.get_or_create_menu(ADMIN_MENU_IDENTIFIER)

        url = admin_reverse(LIST_ALIAS_URL_NAME)
        obj = self.toolbar.get_object()
        language = obj.language if hasattr(obj, "language") else get_language_from_request(self.request)
        if language is None:
            language = get_default_language()
        url += f'?{urlencode({"language": language})}'

        admin_menu.add_sideframe_item(
            _("Aliases"),
            url=url,
            position=self.get_insert_position(admin_menu, self.plural_name),
        )

    def add_alias_menu(self):
        alias_menu = self.toolbar.get_or_create_menu(
            ALIAS_MENU_IDENTIFIER,
            self.name,
            position=1,
        )

        can_change = self.request.user.has_perm(
            get_model_permission_codename(Alias, 'change'),
        )
        alias_menu.add_modal_item(
            _('Change alias settings'),
            url=admin_reverse(
                'djangocms_alias_alias_change',
                args=[self.toolbar.obj.alias_id],
            ),
            disabled=not can_change,
        )
        alias_menu.add_modal_item(
            _('View usage'),
            url=admin_reverse(
                USAGE_ALIAS_URL_NAME,
                args=[self.toolbar.obj.alias_id],
            ),
        )

        # Only show deletion if versioning is not enabled
        if not is_versioning_enabled():
            alias_menu.add_modal_item(
                _('Delete alias'),
                url=admin_reverse(
                    DELETE_ALIAS_URL_NAME,
                    args=(self.toolbar.obj.alias_id, ),
                ),
                on_close=admin_reverse(
                    LIST_ALIAS_URL_NAME,
                ),
                disabled=not can_change,
            )

    @classmethod
    def get_insert_position(cls, admin_menu, item_name):
        """
        Ensures that there is a SHORTCUTS_BREAK and returns a position for an
        alphabetical position against all items between SHORTCUTS_BREAK, and
        the ADMINISTRATION_BREAK.
        """
        start = admin_menu.find_first(Break, identifier=SHORTCUTS_BREAK)

        if not start:
            end = admin_menu.find_first(Break, identifier=ADMINISTRATION_BREAK)
            admin_menu.add_break(SHORTCUTS_BREAK, position=end.index)
            start = admin_menu.find_first(Break, identifier=SHORTCUTS_BREAK)
        end = admin_menu.find_first(Break, identifier=ADMINISTRATION_BREAK)

        items = admin_menu.get_items()[start.index + 1: end.index]
        for idx, item in enumerate(items):
            try:
                if force_str(item_name.lower()) < force_str(item.name.lower()):  # noqa: E501
                    return idx + start.index + 1
            except AttributeError:
                # Some item types do not have a 'name' attribute.
                pass
        return end.index

    def enable_create_wizard_button(self):
        button_lists = [
            result.item
            for result in self.toolbar.find_items(item_type=ButtonList)
        ]
        buttons = list(
            # flatten the list
            itertools.chain.from_iterable([
                item.buttons
                for item in button_lists
            ])
        )

        # There will always be this button, because we are in the context of
        # alias app views
        create_wizard_button = [
            button for button in buttons if button.name == gettext('Create')
        ][0]

        from cms.wizards.wizard_pool import entry_choices

        # we enable this button when user has permissions to perform actions on
        # wizard
        enable_create_wizard_button = bool(
            # entry_choices gets required argument page
            list(entry_choices(self.request.user, page=None))
        )
        create_wizard_button.disabled = not enable_create_wizard_button

    def override_language_switcher(self):
        language_menu = self.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER, _('Language'))
        if not language_menu:
            return
        # Remove all existing language links
        # remove_item uses `items` attribute so we have to copy object
        for _item in copy(language_menu.items):
            language_menu.remove_item(item=_item)

        for code, name in get_language_tuple(self.current_site.pk):
            # Showing only existing translation. For versioning it will be the
            # latest translation i.e. draft or published
            alias_content = self.toolbar.obj.alias.get_content(
                language=code,
                show_draft_content=True,
            )
            if alias_content:
                with force_language(code):
                    url = alias_content.get_absolute_url()
                language_menu.add_link_item(name, url=url, active=self.current_lang == code)

    def change_language_menu(self):
        if self.toolbar.edit_mode_active and isinstance(self.toolbar.obj, AliasContent):
            can_change = self.request.user.has_perm(
                get_model_permission_codename(Alias, 'change'),
            )
        else:
            can_change = False

        if can_change:
            alias_content = self.toolbar.obj
            language_menu = self.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
            if not language_menu:
                return None

            languages = get_language_dict(self.current_site.pk)
            current_placeholder = alias_content.placeholder

            remove = [
                (code, languages.get(code, code))
                for code in alias_content.alias.get_languages()
                if code in languages
            ]
            add = [code for code in languages.items() if code not in remove]
            copy = [
                (code, name)
                for code, name in languages.items()
                if code != self.current_lang and (code, name) in remove and current_placeholder
            ]

            if add or remove or copy:
                language_menu.add_break(ALIAS_LANGUAGE_BREAK)

            if add:
                add_plugins_menu = language_menu.get_or_create_menu(
                    f'{LANGUAGE_MENU_IDENTIFIER}-add',
                    _('Add Translation'),
                )
                add_url = admin_reverse('djangocms_alias_aliascontent_add')

                for code, name in add:
                    url = add_url_parameters(add_url, language=code, alias=alias_content.alias_id)
                    add_plugins_menu.add_modal_item(name, url=url)

            if remove:
                remove_plugins_menu = language_menu.get_or_create_menu(
                    f'{LANGUAGE_MENU_IDENTIFIER}-del',
                    _('Delete Translation'),
                )
                disabled = len(remove) == 1
                for code, name in remove:
                    alias_content = alias_content.alias.get_content(
                        language=code,
                        show_draft_content=True,
                    )
                    translation_delete_url = admin_reverse(
                        'djangocms_alias_aliascontent_delete',
                        args=(alias_content.pk,),
                    )
                    url = add_url_parameters(translation_delete_url, language=code)
                    remove_plugins_menu.add_modal_item(name, url=url, disabled=disabled)

            if copy:
                copy_plugins_menu = language_menu.get_or_create_menu(
                    f'{LANGUAGE_MENU_IDENTIFIER}-copy',
                    _('Copy all plugins')
                )
                title = _('from %s')
                question = _('Are you sure you want to copy all plugins from %s?')

                try:
                    copy_url = admin_reverse('cms_placeholder_copy_plugins')
                except NoReverseMatch:
                    copy_url = admin_reverse('djangocms_alias_alias_copy_plugins')

                for code, name in copy:
                    source_placeholder = alias_content.alias.get_placeholder(
                        language=code,
                        show_draft_content=True,
                    )
                    copy_plugins_menu.add_ajax_item(
                        title % name,
                        action=copy_url,
                        data={
                            'source_language': code,
                            'source_placeholder_id': source_placeholder.pk,
                            'target_language': self.current_lang,
                            'target_placeholder_id': current_placeholder.pk,
                        },
                        question=question % name,
                        on_success=self.toolbar.REFRESH_PAGE
                    )



from importlib import reload

from django.conf import settings
from django.test import TestCase, override_settings

from djangocms_alias import cms_config


class ReferenceConfigTestCase(TestCase):
    @classmethod
    def tearDownClass(cls):
        reload(cms_config)
        super().tearDownClass()

    @override_settings(REFERENCES_ALIAS_MODELS_ENABLED=False)
    def test_references_setting_affects_cms_config_false(self):
        reload(cms_config)
        self.assertFalse(cms_config.AliasCMSConfig.djangocms_references_enabled)

    @override_settings(REFERENCES_ALIAS_MODELS_ENABLED=True)
    def test_references_setting_affects_cms_config_true(self):
        reload(cms_config)
        self.assertTrue(cms_config.AliasCMSConfig.djangocms_references_enabled)

    @override_settings()
    def test_references_setting_affects_cms_config_default(self):
        del settings.REFERENCES_ALIAS_MODELS_ENABLED
        reload(cms_config)
        self.assertTrue(cms_config.AliasCMSConfig.djangocms_references_enabled)



from menus.base import Modifier
from menus.menu_pool import menu_pool

from .constants import PLUGIN_URL_NAME_PREFIX
from .models import AliasContent


class AliasDisableMenu(Modifier):
    """Disable menu rendering on alias pages"""

    def modify(self, request, nodes, namespace, root_id, post_cut, breadcrumb):
        if (
            request.toolbar.app_name == PLUGIN_URL_NAME_PREFIX
            or isinstance(request.toolbar.obj, AliasContent)
        ):
            return []
        return nodes


menu_pool.register_modifier(AliasDisableMenu)



from setuptools import find_packages, setup

import djangocms_alias


CLASSIFIERS = [
    'Environment :: Web Environment',
    'Framework :: Django',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries :: Application Frameworks',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Framework :: Django',
    'Framework :: Django :: 3.2',
    'Framework :: Django :: 4.0',
    'Framework :: Django :: 4.1',
    'Framework :: Django :: 4.2',
    'Framework :: Django :: 5.0',
    'Framework :: Django CMS :: 4.1',
]

INSTALL_REQUIREMENTS = [
    'Django>=3.2',
    'django-parler>=1.4',
    'django-cms>=4.0rc3',
]

setup(
    name='djangocms-alias',
    author='Divio AG',
    author_email='info@divio.ch',
    maintainer='Django CMS Association and contributors',
    maintainer_email='info@django-cms.org',
    url='https://github.com/django-cms/djangocms-alias',
    license='BSD',
    version=djangocms_alias.__version__,
    description=djangocms_alias.__doc__,
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    platforms=['OS Independent'],
    classifiers=CLASSIFIERS,
    install_requires=INSTALL_REQUIREMENTS,
    packages=find_packages(),
    include_package_data=True,
    test_suite='test_settings.run',
)



from collections import ChainMap

from django import template
from django.utils.translation import get_language

from cms.templatetags.cms_tags import PlaceholderOptions
from cms.toolbar.utils import get_toolbar_from_request
from cms.utils import get_current_site, get_language_from_request
from cms.utils.i18n import get_default_language, get_language_list
from cms.utils.placeholder import validate_placeholder_name
from cms.utils.urlutils import add_url_parameters, admin_reverse

from classytags.arguments import Argument, MultiValueArgument
from classytags.core import Tag

from ..constants import (
    DEFAULT_STATIC_ALIAS_CATEGORY_NAME,
    USAGE_ALIAS_URL_NAME,
)
from ..models import Alias, AliasContent, Category
from ..utils import is_versioning_enabled


register = template.Library()


@register.simple_tag(takes_context=False)
def get_alias_usage_view_url(alias, **kwargs):
    url = admin_reverse(USAGE_ALIAS_URL_NAME, args=[alias.pk])
    return add_url_parameters(url, **ChainMap(kwargs))


@register.filter()
def verbose_name(obj):
    return obj._meta.verbose_name


@register.simple_tag(takes_context=True)
def render_alias(context, instance, editable=False):
    request = context['request']

    toolbar = get_toolbar_from_request(request)
    renderer = toolbar.get_content_renderer()

    editable = editable and renderer._placeholders_are_editable
    source = instance.get_placeholder()

    if source:
        content = renderer.render_placeholder(
            placeholder=source,
            context=context,
            editable=editable,
        )
        return content or ''
    return ''


class StaticAlias(Tag):
    """
    This template node is used to render Alias contents and is designed to be a
    replacement for the CMS Static Placeholder.

    eg: {% static_alias "identifier_text" %}
    eg: {% static_alias "identifier_text" site %}

    Keyword arguments:
    static_code -- the unique identifier of the Alias
    site -- If site is supplied an Alias instance will be created per site.
    """
    name = 'static_alias'
    options = PlaceholderOptions(
        Argument('static_code', resolve=True),
        MultiValueArgument('extra_bits', required=False, resolve=False),
        blocks=[
            ('endstatic_alias', 'nodelist'),
        ],
    )

    def _get_alias(self, request, static_code, extra_bits):
        alias_filter_kwargs = {
            'static_code': static_code,
        }
        # Site
        current_site = get_current_site()
        if 'site' in extra_bits:
            alias_filter_kwargs['site'] = current_site
        else:
            alias_filter_kwargs['site_id__isnull'] = True

        if hasattr(request, "toolbar"):
            # Try getting language from the toolbar first (end and view endpoints)
            language = getattr(request.toolbar.get_object(), "language", None)
            if language not in get_language_list(current_site):
                language = get_language_from_request(request)
        else:
            language = get_language_from_request(request)
        if language is None:
            # Might be on non-cms pages
            language = get_language()

            if language is None:
                language = get_default_language()
        # Try and find an Alias to render
        alias = Alias.objects.filter(**alias_filter_kwargs).first()
        # If there is no alias found we need to create one
        if not alias:

            # If versioning is enabled we can only create the records with a logged-in user / staff member
            if is_versioning_enabled() and not request.user.is_authenticated:
                return None

            # Parler's get_or_create doesn't work well with translations, so we must perform our own get or create
            default_category = Category.objects.filter(translations__name=DEFAULT_STATIC_ALIAS_CATEGORY_NAME).first()
            if not default_category:
                default_category = Category.objects.create(name=DEFAULT_STATIC_ALIAS_CATEGORY_NAME)

            alias_creation_kwargs = {
                'static_code': static_code,
                'creation_method': Alias.CREATION_BY_TEMPLATE
            }
            # Site
            if 'site' in extra_bits:
                alias_creation_kwargs['site'] = current_site

            alias = Alias.objects.create(category=default_category, **alias_creation_kwargs)

        if not AliasContent._base_manager.filter(alias=alias, language=language).exists():
            # Create a first content object if none exists in the given language.
            # If versioning is enabled we can only create the records with a logged-in user / staff member
            if is_versioning_enabled() and not request.user.is_authenticated:
                return None

            # Use base manager since we create version objects ourselves
            alias_content = AliasContent._base_manager.create(
                alias=alias,
                name=static_code,
                language=language,
            )

            if is_versioning_enabled():
                from djangocms_versioning.models import Version

                Version.objects.create(content=alias_content, created_by=request.user)
            alias._content_cache[language] = alias_content

        return alias

    def render_tag(self, context, static_code, extra_bits, nodelist=None):
        request = context.get('request')

        if not static_code or not request:
            # an empty string was passed in or the variable is not available in the context
            if nodelist:
                return nodelist.render(context)
            return ''

        validate_placeholder_name(static_code)

        toolbar = get_toolbar_from_request(request)
        renderer = toolbar.get_content_renderer()
        alias = self._get_alias(request, static_code, extra_bits)

        if not alias:
            return ''

        # Get draft contents in edit or preview mode?
        get_draft_content = False
        if toolbar.edit_mode_active or toolbar.preview_mode_active:
            get_draft_content = True

        language = get_language_from_request(request)
        placeholder = alias.get_placeholder(language=language, show_draft_content=get_draft_content)

        if placeholder:
            content = renderer.render_placeholder(
                placeholder=placeholder,
                context=context,
                nodelist=nodelist,
                use_cache=True,
            )
            return content
        return ''


register.tag(StaticAlias.name, StaticAlias)



import operator
from collections import defaultdict

from django.conf import settings
from django.contrib.sites.models import Site
from django.db import models, transaction
from django.db.models import F, Q
from django.utils.encoding import force_str
from django.utils.functional import cached_property
from django.utils.translation import get_language, gettext_lazy as _

from cms.api import add_plugin
from cms.models import CMSPlugin, Placeholder
from cms.models.fields import PlaceholderRelationField
from cms.models.managers import WithUserMixin
from cms.toolbar.utils import get_object_preview_url
from cms.utils.plugins import copy_plugins_to_placeholder
from cms.utils.urlutils import admin_reverse

from parler.models import TranslatableModel, TranslatedFields

from .constants import CHANGE_CATEGORY_URL_NAME
from .utils import is_versioning_enabled


__all__ = [
    'Category',
    'Alias',
    'AliasContent',
    'AliasPlugin',
]


# Add additional choices through the ``settings.py``.
TEMPLATE_DEFAULT = 'default'


def get_templates():
    choices = [
        (TEMPLATE_DEFAULT, _('Default')),
    ]
    choices += getattr(
        settings,
        'DJANGOCMS_ALIAS_TEMPLATES',
        [],
    )
    return choices


class Category(TranslatableModel):
    translations = TranslatedFields(
        name=models.CharField(
            verbose_name=_('name'),
            max_length=120,
        ),
        meta={'unique_together': [('name', 'language_code')]},
    )

    class Meta:
        verbose_name = _('category')
        verbose_name_plural = _('categories')

    def __str__(self):
        # Be sure to be able to see the category name even if it's not in the current language
        return self.safe_translation_getter("name", any_language=True)

    def get_absolute_url(self):
        """Builds the url to the admin category change view"""
        return admin_reverse(CHANGE_CATEGORY_URL_NAME, args=[self.pk])


class Alias(models.Model):
    CREATION_BY_TEMPLATE = 'template'
    CREATION_BY_CODE = 'code'
    CREATION_METHODS = (
        (CREATION_BY_TEMPLATE, _('by template')),
        (CREATION_BY_CODE, _('by code')),
    )
    creation_method = models.CharField(
        verbose_name=_('creation_method'), choices=CREATION_METHODS,
        default=CREATION_BY_CODE, max_length=20, blank=True,
    )
    category = models.ForeignKey(
        Category,
        verbose_name=_('category'),
        related_name='aliases',
        on_delete=models.PROTECT,
    )
    position = models.PositiveIntegerField(
        verbose_name=_('position'),
        default=0,
    )
    static_code = models.CharField(
        verbose_name=_('static code'),
        max_length=255,
        blank=True,
        null=True,
        help_text=_('To render the alias in templates.')
    )
    site = models.ForeignKey(Site, on_delete=models.CASCADE, null=True, blank=True)

    class Meta:
        verbose_name = _('alias')
        verbose_name_plural = _('aliases')
        ordering = ['position']
        unique_together = (('static_code', 'site'),)  # Only restrict instances that have a site specified

    def __init__(self, *args, **kwargs):
        self._plugins_cache = {}
        self._content_cache = {}
        self._content_languages_cache = []
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    @cached_property
    def name(self):
        """Show alias name for current language"""
        return self.get_name() or ''

    @cached_property
    def is_in_use(self):
        return self.cms_plugins.exists()

    @cached_property
    def objects_using(self):
        objects = set()
        object_ids = defaultdict(set)
        plugins = self.cms_plugins.select_related('placeholder').prefetch_related('placeholder__source')
        for plugin in plugins:
            obj = plugin.placeholder.source
            obj_class_name = obj.__class__.__name__
            if obj_class_name.endswith('Content'):
                attr_name = obj_class_name.replace('Content', '').lower()
                attr_related_model = obj._meta.get_field(attr_name).related_model
                id_attr = getattr(obj, f'{attr_name}_id')
                if id_attr:
                    object_ids[attr_related_model].update([id_attr])
                else:
                    objects.update([obj])
            else:
                objects.update([obj])
        objects.update([
            obj
            for model_class, ids in object_ids.items()
            for obj in model_class.objects.filter(pk__in=ids)
        ])
        return list(objects)

    def get_name(self, language=None):
        content = self.get_content(language, show_draft_content=True)
        name = getattr(content, 'name', f'Alias {self.pk} (No content)')
        if is_versioning_enabled() and content:
            from djangocms_versioning.constants import DRAFT
            version = content.versions.first()

            if version.state == DRAFT:
                return f'{name} (Not published)'

        return name

    def get_absolute_url(self, language=None):
        if is_versioning_enabled():
            from djangocms_versioning.helpers import (
                version_list_url_for_grouper,
            )

            return version_list_url_for_grouper(self)
        content = self.get_content(language=language)
        if content:
            return content.get_absolute_url()

    def get_content(self, language=None, show_draft_content=False):
        if not language:
            language = get_language()

        try:
            return self._content_cache[language]
        except KeyError:
            if show_draft_content and is_versioning_enabled():
                qs = self.contents(manager="admin_manager").latest_content()
            else:
                qs = self.contents.all()
            qs = qs.select_related(
                'alias__category',
            ).prefetch_related(
                'placeholders'
            ).filter(language=language)

            self._content_cache[language] = qs.first()
            return self._content_cache[language]

    def get_placeholder(self, language=None, show_draft_content=False):
        content = self.get_content(language=language, show_draft_content=show_draft_content)
        return getattr(content, 'placeholder', None)

    def get_plugins(self, language=None):
        if not language:
            language = get_language()
        try:
            return self._plugins_cache[language]
        except KeyError:
            placeholder = self.get_placeholder(language)
            plugins = placeholder.get_plugins_list() if placeholder else []
            self._plugins_cache[language] = plugins
            return self._plugins_cache[language]

    def get_languages(self):
        if not self._content_languages_cache:
            queryset = self.contents(manager="admin_manager").current_content()
            self._content_languages_cache = queryset.values_list('language', flat=True)
        return self._content_languages_cache

    def clear_cache(self):
        self._plugins_cache = {}
        self._content_cache = {}
        self._content_languages_cache = []

    @transaction.atomic
    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        self.category.aliases.filter(position__gt=self.position).update(
            position=F('position') - 1,
        )

    def save(self, *args, **kwargs):
        if self._state.adding:
            self.position = self.category.aliases.count()
        return super().save(*args, **kwargs)

    def _set_position(self, position):
        previous_position = self.position

        if previous_position > position:  # moving up
            op = operator.add
            position_range = (position, previous_position)
        else:  # moving down
            op = operator.sub
            position_range = (previous_position, position)

        filters = [
            ~Q(pk=self.pk),
            Q(position__range=position_range),
        ]

        self.position = position
        self.save()
        self.category.aliases.filter(*filters).update(position=op(F('position'), 1))  # noqa: E501


class AliasContentManager(WithUserMixin, models.Manager):
    """Adds with_user syntax to AliasContent w/o using versioning"""
    pass


class AliasContent(models.Model):
    alias = models.ForeignKey(
        Alias,
        on_delete=models.CASCADE,
        related_name='contents',
    )
    name = models.CharField(
        verbose_name=_('name'),
        max_length=120,
    )
    placeholders = PlaceholderRelationField()
    placeholder_slotname = 'content'
    language = models.CharField(
        max_length=10,
        default=get_language,
    )

    objects = AliasContentManager()

    class Meta:
        verbose_name = _('alias content')
        verbose_name_plural = _('alias contents')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._meta.get_field('language').choices = settings.LANGUAGES

    def __str__(self):
        return f'{self.name} ({self.language})'

    @cached_property
    def placeholder(self):
        try:
            return self.placeholders.get(slot=self.placeholder_slotname)
        except Placeholder.DoesNotExist:
            from cms.utils.placeholder import rescan_placeholders_for_obj
            rescan_placeholders_for_obj(self)
            return self.placeholders.get(slot=self.placeholder_slotname)

    def get_placeholders(self):
        return [self.placeholder]

    def get_absolute_url(self):
        return get_object_preview_url(self)

    def get_template(self):
        return 'djangocms_alias/alias_content.html'

    @transaction.atomic
    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        self.alias.cms_plugins.filter(language=self.language).delete()

    @transaction.atomic
    def populate(self, replaced_placeholder=None, replaced_plugin=None, plugins=None):
        if not replaced_placeholder and not replaced_plugin:
            copy_plugins_to_placeholder(
                plugins,
                placeholder=self.placeholder,
            )
            return

        if replaced_placeholder:
            plugins = replaced_placeholder.get_plugins(self.language)
            placeholder = replaced_placeholder
            add_plugin_kwargs = {}
        else:
            plugins = CMSPlugin.objects.filter(
                id__in=[replaced_plugin.pk] + replaced_plugin._get_descendants_ids(),
            )
            placeholder = replaced_plugin.placeholder
            add_plugin_kwargs = {'position': 'left', 'target': replaced_plugin}

        copy_plugins_to_placeholder(
            plugins,
            placeholder=self.placeholder,
            language=self.language,
        )
        plugins.delete()
        placeholder._recalculate_plugin_positions(self.language)

        new_plugin = add_plugin(
            placeholder,
            plugin_type='Alias',
            language=self.language,
            alias=self.alias,
            **add_plugin_kwargs
        )
        if replaced_plugin:
            new_plugin.position = replaced_plugin.position
            new_plugin.save(update_fields=['position'])
        return new_plugin


def copy_alias_content(original_content):
    """Copy the AliasContent object and deepcopy its
    placeholders and plugins

    This is needed for versioning integration.
    """
    # Copy content object
    content_fields = {
        field.name: getattr(original_content, field.name)
        for field in AliasContent._meta.fields
        # don't copy primary key because we're creating a new obj
        if AliasContent._meta.pk.name != field.name
    }
    new_content = AliasContent.objects.create(**content_fields)

    # Copy placeholders
    new_placeholders = []
    for placeholder in original_content.placeholders.all():
        placeholder_fields = {
            field.name: getattr(placeholder, field.name)
            for field in Placeholder._meta.fields
            # don't copy primary key because we're creating a new obj
            # and handle the source field later
            if field.name not in [Placeholder._meta.pk.name, 'source']
        }
        if placeholder.source:
            placeholder_fields['source'] = new_content
        new_placeholder = Placeholder.objects.create(**placeholder_fields)
        # Copy plugins
        placeholder.copy_plugins(new_placeholder)
        new_placeholders.append(new_placeholder)
    new_content.placeholders.add(*new_placeholders)

    return new_content


class AliasPlugin(CMSPlugin):
    alias = models.ForeignKey(
        Alias,
        verbose_name=_('alias'),
        related_name='cms_plugins',
        on_delete=models.CASCADE,
    )
    template = models.CharField(
        verbose_name=_('template'),
        choices=get_templates(),
        default=TEMPLATE_DEFAULT,
        max_length=255,
    )

    class Meta:
        verbose_name = _('alias plugin model')
        verbose_name_plural = _('alias plugin models')

    def __str__(self):
        return force_str(self.alias.name)

    def is_recursive(self, language=None):
        # When versioning is enabled it will only get published content
        # placeholder. If does not exist, then None.
        placeholder = self.alias.get_placeholder(language)

        plugins = AliasPlugin.objects.filter(
            placeholder_id=placeholder,
        )
        plugins = plugins.filter(
            Q(pk=self) | Q(alias__contents__placeholders=placeholder)
        )
        return plugins.exists()



from django.apps import apps
from django.conf import settings

from cms.app_base import CMSAppConfig

from .models import AliasContent, AliasPlugin, copy_alias_content
from .rendering import render_alias_content


try:
    apps.get_app_config('djangocms_internalsearch')
    from .internal_search import AliasContentConfig
except (ImportError, LookupError):
    AliasContentConfig = None

djangocms_versioning_installed = apps.is_installed("djangocms_versioning")


class AliasCMSConfig(CMSAppConfig):
    cms_enabled = True
    cms_toolbar_enabled_models = [(AliasContent, render_alias_content)]
    moderated_models = [AliasContent]

    djangocms_moderation_enabled = getattr(
        settings, 'MODERATING_ALIAS_MODELS_ENABLED', True)
    djangocms_versioning_enabled = getattr(
        settings, 'VERSIONING_ALIAS_MODELS_ENABLED', True) and djangocms_versioning_installed

    if djangocms_versioning_enabled:

        from cms.utils.i18n import get_language_tuple

        from djangocms_versioning.datastructures import VersionableItem

        versioning = [
            VersionableItem(
                content_model=AliasContent,
                grouper_field_name='alias',
                extra_grouping_fields=["language"],
                version_list_filter_lookups={"language": get_language_tuple},
                copy_function=copy_alias_content,
                grouper_selector_option_label=lambda obj, lang: obj.get_name(lang),
            ),
        ]

    djangocms_references_enabled = getattr(
        settings, 'REFERENCES_ALIAS_MODELS_ENABLED', True)
    reference_fields = [
        (AliasPlugin, 'alias'),
    ]

    # Internalsearch configuration
    if AliasContentConfig:
        djangocms_internalsearch_enabled = True
        internalsearch_config_list = [
            AliasContentConfig,
        ]



# Generated by Django 2.2.13 on 2020-07-23 14:24

from django.db import migrations, models
import django.db.models.deletion
import parler.fields


class Migration(migrations.Migration):

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
        ('djangocms_alias', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='alias',
            name='creation_method',
            field=models.CharField(blank=True, choices=[('template', 'by template'), ('code', 'by code')], default='code', max_length=20, verbose_name='creation_method'),
        ),
        migrations.AddField(
            model_name='alias',
            name='site',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='sites.Site'),
        ),
        migrations.AddField(
            model_name='alias',
            name='static_code',
            field=models.CharField(blank=True, help_text='To render the alias in templates.', max_length=255, null=True, verbose_name='static code'),
        ),
        migrations.AlterField(
            model_name='categorytranslation',
            name='master',
            field=parler.fields.TranslationsForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='translations', to='djangocms_alias.Category'),
        ),
        migrations.AlterUniqueTogether(
            name='alias',
            unique_together={('static_code', 'site')},
        ),
    ]



from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from .models import Text


@plugin_pool.register_plugin
class TextPlugin(CMSPluginBase):
    model = Text
    name = 'Text'
    allow_children = True
    render_template = 'text/text.html'



from distutils.version import LooseVersion

from django import get_version
from django.contrib.sites.models import Site
from django.core.exceptions import ValidationError
from django.urls import reverse

from cms.api import add_plugin, create_title
from cms.models import Placeholder

from djangocms_alias.cms_plugins import Alias
from djangocms_alias.models import Alias as AliasModel, AliasContent, Category
from djangocms_alias.utils import is_versioning_enabled

from .base import BaseAliasPluginTestCase


DJANGO_VERSION = get_version()
DJANGO_4_0 = LooseVersion(DJANGO_VERSION) < LooseVersion('4.1')


class AliasModelsTestCase(BaseAliasPluginTestCase):

    def _get_aliases_positions(self, category):
        return dict(category.aliases.values_list('position', 'pk'))

    def test_alias_placeholder_slot_save_again(self):
        alias = self._create_alias(self.placeholder.get_plugins())
        alias_placeholder = alias.get_placeholder(self.language)
        slot_name = alias_placeholder.slot
        alias.save()
        self.assertEqual(alias_placeholder.slot, slot_name)

    def test_alias_placeholder_name(self):
        alias = self._create_alias(
            self.placeholder.get_plugins(),
            name='test alias 2',
        )
        self.assertEqual(str(alias), 'test alias 2')

    def test_alias_is_not_recursive(self):
        alias = self._create_alias(
            self.placeholder.get_plugins(),
        )
        alias_plugin = add_plugin(
            self.placeholder,
            Alias,
            language=self.language,
            alias=alias,
        )
        self.assertFalse(alias_plugin.is_recursive())

    def test_alias_is_recursive(self):
        alias = self._create_alias(
            self.placeholder.get_plugins(),
        )
        recursed_alias_plugin = add_plugin(
            alias.get_placeholder(self.language),
            Alias,
            language=self.language,
            alias=alias,
        )
        self.assertTrue(recursed_alias_plugin.is_recursive())

    def test_alias_is_recursive_with_custom_template(self):
        alias = self._create_alias(
            self.placeholder.get_plugins(),
        )
        recursive_alias_plugin = add_plugin(
            alias.get_placeholder(self.language),
            Alias,
            language=self.language,
            alias=alias,
            template='custom_alias_template'
        )
        self.assertTrue(recursive_alias_plugin.is_recursive())

    def test_increment_position_for_newly_created_instances(self):
        category1 = Category.objects.create(
            name='Cat 1',
        )
        category2 = Category.objects.create(
            name='Cat 2',
        )
        alias1cat1 = AliasModel.objects.create(category=category1)
        alias1cat2 = AliasModel.objects.create(category=category2)
        alias2cat2 = AliasModel.objects.create(category=category2)
        alias2cat1 = AliasModel.objects.create(category=category1)

        self.assertEqual(
            self._get_aliases_positions(alias1cat1.category),
            {0: alias1cat1.pk, 1: alias2cat1.pk},
        )
        self.assertEqual(
            self._get_aliases_positions(alias1cat2.category),
            {0: alias1cat2.pk, 1: alias2cat2.pk},
        )

    def test_save_and_delete_inc_dec_of_position_delete_first(self):
        alias1 = AliasModel.objects.create(category=self.category)
        alias2 = AliasModel.objects.create(category=self.category)
        alias3 = AliasModel.objects.create(category=self.category)
        alias4 = AliasModel.objects.create(category=self.category)

        alias1.delete()
        self.assertEqual(
            self._get_aliases_positions(alias1.category),
            {0: alias2.pk, 1: alias3.pk, 2: alias4.pk},
        )

    def test_save_and_delete_inc_dec_of_position_delete_in_the_middle(self):
        alias1 = AliasModel.objects.create(category=self.category)
        alias2 = AliasModel.objects.create(category=self.category)
        alias3 = AliasModel.objects.create(category=self.category)

        alias2.delete()
        self.assertEqual(
            self._get_aliases_positions(alias1.category),
            {0: alias1.pk, 1: alias3.pk},
        )

    def test_save_and_delete_inc_dec_of_position_delete_last(self):
        alias1 = AliasModel.objects.create(category=self.category)
        alias2 = AliasModel.objects.create(category=self.category)
        alias3 = AliasModel.objects.create(category=self.category)

        alias3.delete()
        self.assertEqual(
            self._get_aliases_positions(alias1.category),
            {0: alias1.pk, 1: alias2.pk},
        )

    def test_set_position_dont_change_position_in_other_categories(self):
        category1 = Category.objects.create(
            name='Cat 1',
        )
        category2 = Category.objects.create(
            name='Cat 2',
        )
        alias1cat1 = AliasModel.objects.create(category=category1)  # 0
        alias2cat1 = AliasModel.objects.create(category=category1)  # 1
        alias1cat2 = AliasModel.objects.create(category=category2)  # 0
        alias2cat2 = AliasModel.objects.create(category=category2)  # 1

        alias1cat1._set_position(alias2cat1.position)
        self.assertEqual(
            self._get_aliases_positions(alias1cat1.category),
            {0: alias2cat1.pk, 1: alias1cat1.pk},
        )
        self.assertEqual(
            self._get_aliases_positions(alias1cat2.category),
            {0: alias1cat2.pk, 1: alias2cat2.pk},
        )

    def test_set_position_moving_up(self):
        alias1 = AliasModel.objects.create(category=self.category)  # 0
        alias2 = AliasModel.objects.create(category=self.category)  # 1
        alias3 = AliasModel.objects.create(category=self.category)  # 2
        alias4 = AliasModel.objects.create(category=self.category)  # 3

        alias3._set_position(1)
        self.assertEqual(
            self._get_aliases_positions(alias1.category),
            {0: alias1.pk, 1: alias3.pk, 2: alias2.pk, 3: alias4.pk},
        )

    def test_set_position_moving_down(self):
        alias1 = AliasModel.objects.create(category=self.category)  # 0
        alias2 = AliasModel.objects.create(category=self.category)  # 1
        alias3 = AliasModel.objects.create(category=self.category)  # 2
        alias4 = AliasModel.objects.create(category=self.category)  # 3

        alias2._set_position(2)
        self.assertEqual(
            self._get_aliases_positions(alias1.category),
            {0: alias1.pk, 1: alias3.pk, 2: alias2.pk, 3: alias4.pk},
        )

    def test_set_position_moving_up_from_end_to_start(self):
        alias1 = AliasModel.objects.create(category=self.category)  # 0
        alias2 = AliasModel.objects.create(category=self.category)  # 1
        alias3 = AliasModel.objects.create(category=self.category)  # 2
        alias4 = AliasModel.objects.create(category=self.category)  # 3

        alias4._set_position(0)
        self.assertEqual(
            self._get_aliases_positions(alias1.category),
            {0: alias4.pk, 1: alias1.pk, 2: alias2.pk, 3: alias3.pk},
        )

    def test_set_position_moving_down_from_start_to_end(self):
        alias1 = AliasModel.objects.create(category=self.category)  # 0
        alias2 = AliasModel.objects.create(category=self.category)  # 1
        alias3 = AliasModel.objects.create(category=self.category)  # 2
        alias4 = AliasModel.objects.create(category=self.category)  # 3

        alias1._set_position(3)
        self.assertEqual(
            self._get_aliases_positions(alias1.category),
            {0: alias2.pk, 1: alias3.pk, 2: alias4.pk, 3: alias1.pk},
        )

    def test_pages_using_alias(self):
        site1 = Site.objects.create(domain='site1.com', name='1')
        site2 = Site.objects.create(domain='site2.com', name='2')
        alias = self._create_alias(name='alias')

        site1_page = self._create_page(
            title='Site1',
            language=self.language,
            site=site1,
        )
        self.add_alias_plugin_to_page(site1_page, alias)
        # Should show on the list

        nested_page1 = self._create_page(
            title='Site1 nested page 1',
            language=self.language,
            site=site1,
            parent=site1_page,
        )
        self.add_alias_plugin_to_page(nested_page1, alias)
        self.add_alias_plugin_to_page(nested_page1, alias)
        # Should show on the list only once

        nested_page2 = self._create_page(
            title='Site1 nested page 2',
            language=self.language,
            site=site1,
            parent=site1_page,
        )
        self.add_alias_plugin_to_page(nested_page2, alias)
        # Should show on the list

        nested_page3 = self._create_page(
            title='Site1 nested page 3',
            language=self.language,
            site=site1,
            parent=site1_page,
        )  # Not show on the list

        deep_nested_page4 = self._create_page(
            title='Site1 deep nested page 4',
            language=self.language,
            site=site1,
            parent=nested_page3,
        )
        self.add_alias_plugin_to_page(deep_nested_page4, alias)
        # Should show on the list

        site2_page = self._create_page(
            title='Site2',
            language='de',
            site=site2,
        )
        self.add_alias_plugin_to_page(site2_page, alias, 'de')

        if is_versioning_enabled():
            create_title('en', 'Site2 EN', site2_page, created_by=self.superuser)
            self._publish(site2_page, 'en')
        else:
            create_title('en', 'Site2 EN', site2_page)

        self.add_alias_plugin_to_page(site2_page, alias, 'en')
        # Should show on the list only once

        with self.assertNumQueries(3):
            objects = alias.objects_using

        self.assertEqual(
            sorted(obj.pk for obj in objects),
            [
                site1_page.pk,
                nested_page1.pk,
                nested_page2.pk,
                deep_nested_page4.pk,
                site2_page.pk,
            ]
        )

    def test_aliases_using_alias(self):
        root_alias = self._create_alias(name='root alias')
        if not is_versioning_enabled():
            # TODO: fix it after versioning will have multilanguage support
            AliasContent.objects.create(
                name='root alias de',
                alias=root_alias,
                language='de',
            )
            AliasContent.objects.create(
                name='root alias it',
                alias=root_alias,
                language='it',
            )
        root_alias2 = self._create_alias(name='root alias 2')

        alias1 = self._create_alias(name='alias 1')
        alias2 = self._create_alias(name='alias 2')
        alias3 = self._create_alias(name='alias 3')
        alias4 = self._create_alias(name='alias 4')

        add_plugin(
            root_alias.get_placeholder(self.language),
            'Alias',
            language=self.language,
            alias=alias1,
        )
        if not is_versioning_enabled():
            # TODO: fix it after versioning will have multilanguage support
            add_plugin(
                root_alias.get_placeholder('de'),
                'Alias',
                language='de',
                alias=alias1,
            )
            add_plugin(
                root_alias.get_placeholder('it'),
                'Alias',
                language='it',
                alias=alias1,
            )
        # Alias1 should show only once
        add_plugin(
            root_alias.get_placeholder(self.language),
            'Alias',
            language=self.language,
            alias=alias2,
        )
        add_plugin(
            root_alias2.get_placeholder(self.language),
            'Alias',
            language=self.language,
            alias=alias2,
        )
        add_plugin(
            alias2.get_placeholder(self.language),
            'Alias',
            language=self.language,
            alias=alias3,
        )
        add_plugin(
            alias3.get_placeholder(self.language),
            'Alias',
            language=self.language,
            alias=alias4,
        )

        with self.assertNumQueries(3):
            objects = alias1.objects_using
        self.assertEqual(
            sorted(obj.pk for obj in objects),
            [root_alias.pk],
        )

        with self.assertNumQueries(3):
            objects = alias2.objects_using
        self.assertEqual(
            sorted(obj.pk for obj in objects),
            [root_alias.pk, root_alias2.pk],
        )

        with self.assertNumQueries(3):
            objects = alias3.objects_using
        self.assertEqual(
            sorted(obj.pk for obj in objects),
            [alias2.pk],
        )

        with self.assertNumQueries(3):
            objects = alias4.objects_using
        self.assertEqual(
            sorted(obj.pk for obj in objects),
            [alias3.pk],
        )

    def test_pages_and_aliases_using_objects(self):
        alias = self._create_alias()
        root_alias = self._create_alias(name='root alias')
        add_plugin(
            root_alias.get_placeholder(self.language),
            'Alias',
            language=self.language,
            alias=alias,
        )
        self.add_alias_plugin_to_page(self.page, alias, 'en')
        with self.assertNumQueries(5):
            objects = alias.objects_using
        self.assertEqual(
            sorted(obj.pk for obj in objects),
            [self.page.pk, root_alias.pk],
        )

    def test_delete(self):
        self.page.delete()

        alias = self._create_alias([self.plugin])
        add_plugin(
            self.placeholder,
            Alias,
            language=self.language,
            alias=alias,
        )
        self.assertEqual(Placeholder.objects.count(), 1)
        alias.delete()
        self.assertFalse(alias.__class__.objects.filter(pk=alias.pk).exists())
        if DJANGO_4_0:
            self.assertEqual(alias.cms_plugins.count(), 0)
        self.assertEqual(Placeholder.objects.count(), 0)
        alias.save()  # Django 4.1+ disallows to use relations (cmsplugins) of unsaved objects.
        self.assertEqual(alias.cms_plugins.count(), 0)

    def test_category_get_absolute_url(self):
        """
        Category uses the admin change view as its absolute url
        """
        category = Category.objects.create(name="Test Category")

        app_label = category._meta.app_label
        expected = reverse(
            f"admin:{app_label}_category_change", args=[category.pk]
        )

        self.assertEqual(category.get_absolute_url(), expected)

    def test_category_name_same_across_languages(self):
        """
        Category name may be the same across languages
        """
        category = Category.objects.create(name='Samename A')
        category.set_current_language('de')
        category.name = "Samename A"
        try:
            category.validate_unique()
        except ValidationError:
            self.fail("Same Category name should be allowed across two languages.")

        category.set_current_language('en')
        self.assertEqual(category.name, 'Samename A')
        category.set_current_language('de')
        self.assertEqual(category.name, 'Samename A')

    def test_category_name_unique_for_language(self):
        """
        Category name can't be the same in one language for two different categories
        """
        with self.login_user_context(self.superuser):
            Category.objects.create(name='Samename B')
            c = Category(name='Samename B')
            self.assertRaises(ValidationError, c.validate_unique)
